import yaml, collections, sys, os, glob

y=yaml.load(open(sys.argv[1]), Loader=yaml.CSafeLoader)
r=y["report"]

if sys.argv[1].endswith("-binaries.yml"):
    t="depend"
else:
    t="build-depend"

by_missing = collections.defaultdict(list)


def map_reason(unsat):
  try:
    name, arch = unsat.split()[0].split(":")
  except:
    name, arch = unsat.split()[0], "native"

  reason = "src:" + sources[name] if name in sources else name

  if name.startswith("python3-") and False:
     return (3, "python3-*")
  if 'python' in name and False:
    return (3, "*python*")
  if name.startswith("python-") and False:
     return (3, "python-*")
  if name.startswith("libghc"):
     return (3, "libghc*")
  if name.startswith("librust"):
     return (3, "librust*")
  if name.startswith("golang-"):
     return (3, "golang-*")
  if name.startswith("node-"):
     return (3, "node-*")
  if name.startswith("ruby-"):
     return (3, "ruby-*")

  return (3, reason)

# Build set of source packages in the release pocket
in_release = None
tasks = {}
sources = {}

import apt_pkg
for file in glob.glob("root/var/lib/apt/lists/*Packages"):
    with apt_pkg.TagFile(file) as tags:
        for section in tags:
            try:
                tasks[section["Package"]] = set(s.strip() for s in section["Task"].split(","))
            except KeyError:
                continue

for file in glob.glob("root/var/lib/apt/lists/*Sources"):
    with apt_pkg.TagFile(file) as tags:
        for section in tags:
            src = section["Package"]
            try:
                for line in section["Package-List"].splitlines():
                    if line.strip():
                        binary = line.split()[0]
                        sources[binary] = src
            except KeyError:
                for binary in (s.strip() for s in section["Binary"].split(",")):
                    sources[binary] = src

assert "python3-gdbm" in sources

if "armhf" in os.getcwd():
    import apt_pkg, os

    apt_pkg.config["dir"] = os.path.abspath("root")
    apt_pkg.config["apt::architecture"] = "armhf"
    apt_pkg.init()

    cache = apt_pkg.Cache(progress=None)
    pkgrecords = apt_pkg.PackageRecords(cache)

    in_release = set()
    for package in cache.packages:
        for version in package.version_list:
            # Filter out all
            if version.arch != package.architecture:
                continue
            for file, offset in version.file_list:
                if "-" in file.archive: # looking for release pocket
                    continue
                if t == "build-depend":
                    pkgrecords.lookup((file, offset))
                    in_release.add(pkgrecords.source_pkg)
                else:
                    in_release.add(package.name)

print("Unsatisfiable %ss (%s)" % (t, sys.argv[2]))
print("=======================================" + "=" * len(sys.argv[2]))
print("As of release Release file:", y["date"])
print("As of proposed Release file:", y["date-proposed"])
if "--only-regressions" in sys.argv and in_release:
    print("Only showing migration issues, that is, where the source package has %s packages in release pocket" % apt_pkg.config["apt::architecture"])
print()

conflicts=[]
for p in r:
    if in_release and p["package"] not in in_release and "--only-regressions" in sys.argv:
        continue
    not_a_regression = ", but not a regression" if in_release and p["package"] not in in_release else ""
    for reason in p["reasons"]:
        if "conflict" in reason:
            s = ("%s " + t + "s on conflicting packages: %s=%s vs %s=%s%s") % (p["package"], reason["conflict"]["pkg1"]["package"], reason["conflict"]["pkg1"]["version"], reason["conflict"]["pkg2"]["package"], reason["conflict"]["pkg2"]["version"], not_a_regression)
            if "depchain1" in reason["conflict"]:
                #s += "\n    - Dependency chain 1:\n      " + "\n      ".join(yaml.safe_dump(reason["conflict"]["depchain1"], sort_keys=False).split("\n"))
                s += "\n    - chain 1: " + ", ".join(" -> ".join(item["depends"]  for item in chain["depchain"]) for chain in reason["conflict"]["depchain1"])
            if "depchain2" in reason["conflict"]:
                #s += "\n    - Dependency chain 2:\n      " + "\n      ".join(yaml.safe_dump(reason["conflict"]["depchain2"], sort_keys=False).split("\n"))
                s += "\n    - chain 2: " + ", ".join(" -> ".join(item["depends"]  for item in chain["depchain"]) for chain in reason["conflict"]["depchain2"])
            conflicts.append(s)


        if "missing" in reason:
            unsat = reason["missing"]["pkg"]["unsat-dependency"]

            if unsat.startswith("linux-restricted-generate"):
                continue

            chains = []
            for chain in reason["missing"].get("depchains", []):
                chains.append(" -> ".join(item["depends"]  for item in chain["depchain"]))

            if p["package"] in tasks and t == "depend":
                for task in sorted(tasks[p["package"]]):
                    by_missing[map_reason(unsat)].append(("%-40s " + t + "s on missing %s%s%s%s") % ("task:" + task, unsat, " via " + p["package"], " and " + ", ".join(chains) if chains else "", not_a_regression))
            by_missing[map_reason(unsat)].append(("%-40s " + t + "s on missing %s%s%s") % (p["package"], unsat, " via " + ", ".join(chains) if chains else "", not_a_regression))

print("Trivial conflicting %ss in the same package (probably hardcoded dependency vs shlibs): " % t)
print("  * " + "\n\n  * ".join(sorted(c for c in conflicts if "->" not in c)))
print()

print("Summary:")

for prio, reason in sorted(by_missing, key=lambda k: (k[0], -len(by_missing[k]), k[1])):
    if reason == "": continue
    print("  * {} packages or tasks {}ing on {}".format(len(sorted(by_missing[prio, reason])), t, reason))

print()

for prio, reason in sorted(by_missing, key=lambda k: (k[0], -len(by_missing[k]), k[1])):
    if reason == "": continue
    print(("Packages " + t + "ing on %s:") % reason)
    print("  * " + "\n  * ".join(sorted(by_missing[prio, reason])))
    print()

print("Other Conflicts: ")
print("  * " + "\n\n  * ".join(sorted(c for c in conflicts if "->" in c)))
